#!/usr/bin/env python3
"""
The TinyPseuCoParser.
It attempts to parse a TinyPseuCo program, and can simplify it into a μPseuCo one
"""

# pylint: disable=line-too-long
# pylint: disable=protected-access

import enum
from enum import Enum
from collections import defaultdict, deque
import itertools

import typing
from typing import DefaultDict, Deque, List, Union, Any, Mapping

import abc

import pyparsing as pp
from pyparsing import (oneOf, opAssoc, Word, infixNotation, nums, alphas,
                       Literal, Optional, ZeroOrMore, Suppress, delimitedList)

# consider turning packrat on once it is clear that there are no side effects
pp.ParserElement.enablePackrat()


class UnsupportedPseuCoException(pp.ParseFatalException):
    """Exception which is raised when an unsupported part of PseuCo in encountered"""
    def __init__(self, s, loc, message):
        super().__init__(s, loc, "Error: %s" % message)


class ASTNode():
    """basic ASTNode"""

    def __init__(self):
        self.ast_parent: typing.Optional[ASTNode] = None

    def create_label(self):
        """Creates a label to show in the CFG representation"""
        return repr(self)

    def set_parent(self, parent: 'ASTNode'):
        """Sets the nodes parent to :parent: to allow tree traversal"""
        self.ast_parent = parent


@enum.unique
class PseuCoType(Enum):
    """The types supported in TinyPseuCo"""
    String = "string"
    Int = "int"
    Boolean = "bool"

    def get_default_value(self):
        """Returns a default value for each base type"""
        if self is PseuCoType.String:
            return StringLiteral.default_value()
        elif self is PseuCoType.Int:
            return IntegerConstant.default_value()
        elif self is PseuCoType.Boolean:
            return BooleanConstant.default_value()
        raise AssertionError("There shouldn't be another type")


@enum.unique
class BinaryOperator(Enum):
    """Represents all binary operators"""
    Plus = "+"
    Minus = "-"
    Mult = "*"
    Div = "/"
    And = "&&"
    Or = "||"
    Equal = "=="
    NEqual = "!="
    Lt = "<"
    Lte = "<="
    Gt = ">"
    Gte = ">="


@enum.unique
class UnaryOperator(Enum):
    """Represents all unary operators"""
    Plus = "+"
    Minus = "-"
    Not = "!"


class ReplacementCounter():  # pylint: disable=too-few-public-methods
    """Used to assign unique ids to statements after simplification
    Also used to keep count how many global reads/writes occur in a
    function
    """
    def __init__(self, count=0):
        self.count = count
        self.global_operations = 0


class BaseExpression(ASTNode, metaclass=abc.ABCMeta):
    """Base class for expressions in TinyPseuCo"""

    def __init__(self):
        super().__init__()
        self.read_global_variables: typing.List[str] = []

    @property
    @abc.abstractmethod
    def pseuco_type(self) -> PseuCoType:
        """Computes the type of an expression"""

    def extract_global_read(self, tt: "TypeTable", replacement_counter: ReplacementCounter):
        """Replaces global reads with local ones by extracting the global reads
        into temporary variables"""
        extracted = []
        replacements: DefaultDict[str, Deque[str]] = defaultdict(deque)
        for variable in self.read_global_variables:
            var_type = tt.lookup_type(variable)
            new_variable = "$%d" % replacement_counter.count
            replacement_counter.count += 1
            replacements[variable].append(new_variable)
            new_decl = VariableDeclaration(
                None,
                None,
                [  # pylint: disable=abstract-class-instantiated, line-too-long
                    var_type,
                    new_variable,
                    VariableUsage(None, None, [variable], hack_type=var_type, type_table=tt),
                ], type_table=tt)
            replacement_counter.global_operations += 1
            extracted.append(new_decl)
        return extracted, replacements

    @abc.abstractmethod
    def replace_globals(self, replacement):
        """Replaces all occurences of a global variable in an exrpession with a temporary from
        replacement"""
        self.read_global_variables = []


class UnaryExpression(BaseExpression):
    """Unary expression (Not, unary minus, unary plus)"""
    def __init__(self, operator: str, operand) -> None:
        super().__init__()
        self.operator = UnaryOperator(operator)
        self.operand: BaseExpression = operand
        self.operand.set_parent(self)
        self._type = self._computeType()
        self.read_global_variables = self.operand.read_global_variables

    def replace_globals(self, replacement):
        super().replace_globals(replacement)
        self.operand.replace_globals(replacement)

    @property
    def pseuco_type(self) -> PseuCoType:
        return self._type

    def _computeType(self):
        """Computes the type of the expression"""
        operand_t = self.operand.pseuco_type
        if self.operator is UnaryOperator.Plus or self.operator is UnaryOperator.Minus:
            assert operand_t is PseuCoType.Int, "operator only works on integers"
            return PseuCoType.Int
        elif self.operator is UnaryOperator.Not:
            assert operand_t is PseuCoType.Boolean, "not only works on booleans"
            return PseuCoType.Boolean
        else:
            raise NotImplementedError("unknown operator")

    def __repr__(self):
        return "({} {})".format(self.operator.value, self.operand)


class BinaryExpression(BaseExpression):
    """Binary expression for arithmetic operators, logical connectives and
    comparision operators"""
    def __init__(self, operator: str, lhs, rhs) -> None:
        super().__init__()
        self.operator: BinaryOperator = BinaryOperator(operator)
        self.lhs: BaseExpression = lhs
        self.rhs: BaseExpression = rhs
        self.lhs.set_parent(self)
        self.rhs.set_parent(self)
        self._type = self._computeType()
        self.read_global_variables = (
            self.lhs.read_global_variables + self.rhs.read_global_variables)

    def __repr__(self):
        return "({} {} {})".format(self.lhs, self.operator.value, self.rhs)

    def replace_globals(self, replacement):
        super().replace_globals(replacement)
        self.lhs.replace_globals(replacement)
        self.rhs.replace_globals(replacement)

    @property
    def pseuco_type(self) -> PseuCoType:
        return self._type

    def _computeType(self) -> PseuCoType:
        lhs = self.lhs.pseuco_type
        rhs = self.rhs.pseuco_type
        if self.operator is BinaryOperator.Equal or self.operator is BinaryOperator.NEqual:
            assert lhs == rhs, "Types must match for comparision"
            return PseuCoType.Boolean
        elif self.operator is BinaryOperator.And or self.operator is BinaryOperator.Or:
            assert lhs is PseuCoType.Boolean, "lhs must be boolean"
            assert lhs == rhs, "rhs must be boolean"
            return PseuCoType.Boolean
        elif any((self.operator is op
                  for op in (BinaryOperator.Lt, BinaryOperator.Lte,
                             BinaryOperator.Gt, BinaryOperator.Gte))):
            assert lhs is PseuCoType.Int, "lhs must be integer"
            assert lhs == rhs, "rhs must be integer"
            return PseuCoType.Boolean
        elif any((self.operator is op
                  for op in (BinaryOperator.Minus, BinaryOperator.Mult,
                             BinaryOperator.Div))):
            assert lhs is PseuCoType.Int, "lhs must be integer"
            assert lhs == rhs, "rhs must be integer"
            return PseuCoType.Int
        elif self.operator is BinaryOperator.Plus:
            if lhs is PseuCoType.String:
                assert rhs is PseuCoType.Boolean,\
                    "cannot concatenate string and bool"
                return PseuCoType.String
            assert lhs is PseuCoType.Int, "lhs must be integer"
            assert lhs == rhs, "rhs must be integer"
            return PseuCoType.Int
        else:
            raise NotImplementedError(self.operator)


def expression_list2_binary_expression(
        exprlist_: List[List[Union[BinaryExpression, str]]]) -> BinaryExpression:
    """pyparsing returns a list, this converts it into an expression tree"""
    assert len(exprlist_) == 1
    exprlist: List[Union[BinaryExpression, str]] = exprlist_[0]
    assert len(exprlist) >= 3
    assert isinstance(exprlist[1], str)
    operator = exprlist[1]
    operands: List[BinaryExpression] = exprlist[0::2]  # type: ignore
    assert len(operands) >= 2
    lhs = operands.pop(0)
    while operands:
        rhs = operands.pop(0)
        lhs = BinaryExpression(operator, lhs, rhs)
    return lhs


def expression_list2_unary_expression(exprlist_):
    """pyparsing returns a list, this converts it into an expression tree"""
    assert len(exprlist_) == 1
    exprlist = exprlist_[0]
    assert len(exprlist) == 2
    return UnaryExpression(exprlist[0], exprlist[1])


class VariableUsage(BaseExpression):
    """Represents using a variable in an expression"""
    def __init__(self, s, l, tokens, *, type_table, hack_type=None):  # pylint:disable=unused-argument
        super().__init__()
        assert len(tokens) == 1
        self.varname = tokens[0]
        if hack_type is not None:
            # VariableUsage is constructed for extracting global reads
            # instead of directly from the parser as we'd normally do
            self._type = hack_type
            self.read_global_variables.append(self.varname)
        else:
            self._type: PseuCoType = type_table.lookup_type(self.varname)
            if not type_table.is_local_variable(self.varname):
                self.read_global_variables.append(self.varname)

    @property
    def pseuco_type(self) -> PseuCoType:
        return self._type

    def replace_globals(self, replacement):
        super().replace_globals(replacement)
        if self.varname in replacement:
            self.varname = replacement[self.varname].popleft()

    def __repr__(self):
        return self.varname


class StringLiteral(BaseExpression):
    """Represents a string literal"""
    def __init__(self, tokens):
        super().__init__()
        assert len(tokens) == 1
        self.text = tokens[0]

    def replace_globals(self, _):
        pass

    def __repr__(self):
        return '"{}"'.format(self.text)

    @staticmethod
    def default_value():
        """Returns the default value for a string (used in variable initialization)"""
        return StringLiteral([""])

    @property
    def pseuco_type(self) -> PseuCoType:
        return PseuCoType.String


class BooleanConstant(BaseExpression):
    """Represents a boolean constant"""
    def __init__(self, tokens):
        super().__init__()
        assert len(tokens) == 1
        self.value = tokens[0] == "true"

    def replace_globals(self, _):
        pass

    def __repr__(self):
        return '{}'.format(str(self.value).lower())

    @staticmethod
    def default_value():
        """Returns the default boolean value (used in variable initialization)"""
        return BooleanConstant(["False"])

    @property
    def pseuco_type(self) -> PseuCoType:
        return PseuCoType.Boolean


class IntegerConstant(BaseExpression):
    """Represents an integer constant"""
    def __init__(self, tokens):
        super().__init__()
        assert len(tokens) == 1
        self.value = int(tokens[0])

    def replace_globals(self, _):
        pass

    def __repr__(self):
        return '{}'.format(self.value)

    @property
    def pseuco_type(self) -> PseuCoType:
        return PseuCoType.Int

    @staticmethod
    def default_value():
        """Returns the default integer value (used in variable initialization)"""
        return IntegerConstant(["0"])


class Statement(ASTNode, metaclass=abc.ABCMeta):
    """The base class of all statements"""
    def __init__(self):
        super().__init__()
        self.node_classes: typing.List[str] = ["normal"]
        self.node_shape: str = "rect"
        self.id = 0

    def _set_id(self, id_: int) -> int:
        self.id = id_
        return id_ + 1

    @abc.abstractmethod
    def get_cfg_exit_nodes(self) -> List["Statement"]:
        """Returns a list of all exit nodes for the CFG construction"""
        return [self]

    @abc.abstractmethod
    def simplify(self, tt: "TypeTable", replacement_counter: ReplacementCounter) -> List["Statement"]:
        """Converts a TinyPseuCo statement into a μPseuCo one"""

    @abc.abstractmethod
    def create_node_data(self, cfg_predecessors: List["Statement"]) -> List[Mapping[str, Any]]:
        """Creates a serializable representation of the node to send to the JS world"""
        ret = {
            "id": self.id,
            "node_label": self.create_label(),
            "node_class": " ".join(self.node_classes),
            "node_shape": self.node_shape,
            "parents":  [node.id for node in cfg_predecessors]
        }
        return [ret]


class EmptyStatement(Statement):
    """The empty statement. Can occur in an ITE or when someone presses ; twice by accident"""
    def __init__(self, tokens):
        super().__init__()
        assert not tokens

    def __repr__(self):
        return ";"

    def simplify(self, tt, replacement_counter):
        return [self]

    def get_cfg_exit_nodes(self):  # pylint: disable=useless-super-delegation
        return super().get_cfg_exit_nodes()

    def create_node_data(self, cfg_predecessors):  # pylint: disable=useless-super-delegation
        return super().create_node_data(cfg_predecessors)


class SimpleStatement(Statement):
    """A statement consisting solely of an expression"""
    def __init__(self, tokens):
        super().__init__()
        assert len(tokens) == 1
        self.expression: BaseExpression = tokens[0]

    def __repr__(self):
        return repr(self.expression) + ";"

    def get_cfg_exit_nodes(self):  # pylint: disable=useless-super-delegation
        return super().get_cfg_exit_nodes()

    def create_node_data(self, cfg_predecessors):  # pylint: disable=useless-super-delegation
        return super().create_node_data(cfg_predecessors)

    def simplify(self, tt, replacement_counter):
        reads, replacements = self.expression.extract_global_read(tt, replacement_counter)
        self.expression.replace_globals(replacements, replacement_counter)
        reads.append(self)
        return reads


class MultiStatement(Statement):
    """MultiStatement: a sequence of multiple statements (unsupported: opens a new scope)"""
    def __init__(self, tokens):
        super().__init__()
        self.statements = tokens

    def simplify(self, tt, replacement_counter):
        simplified = (statement.simplify(tt, replacement_counter) for statement in self.statements)
        return [MultiStatement(list(itertools.chain.from_iterable(simplified)))]

    def _set_id(self, id_):
        for statement in self.statements:
            id_ = statement._set_id(id_)
        if not self.statements:
            id_ = super()._set_id(id_)
        return id_

    def __repr__(self):
        return "{{{0}}}".format("\n".join(
            repr(s) for s in self.statements))

    def get_cfg_exit_nodes(self):
        if self.statements:
            return self.statements[-1].get_cfg_exit_nodes()
        return [self]

    def create_node_data(self, cfg_predecessors):
        if not self.statements:
            return super().create_node_data(cfg_predecessors)
        ret = []
        for statement in self.statements:
            ret += statement.create_node_data(cfg_predecessors)
            cfg_predecessors = statement.get_cfg_exit_nodes()
        return ret


class SimpleAgent(ASTNode):
    """Represents an agent declaration (takes no arguments, returns void)"""
    def __init__(self, s, l, tokens, *, type_table):  # pylint: disable=unused-argument
        super().__init__()
        self.agentName, functionBody, *empty = tokens
        self.local_environment = type_table.close_fresh_environment()
        statements = []
        replacement_counter = ReplacementCounter(0)
        for statement in functionBody.statements:
            statements.extend(statement.simplify(type_table, replacement_counter))
        self.global_operations = replacement_counter.global_operations
        id_ = 0
        for statement in statements:
            id_ = statement._set_id(id_)
        functionBody.statements = statements
        self.functionBody = functionBody
        assert [] == empty

    def __repr__(self):
        ret = "void " + self.agentName + "() {\n" + repr(
            self.functionBody) + "\n}"
        return ret


class FunctionBody(ASTNode):
    """The body of a function. Mostly a list of statements"""
    def __init__(self, tokens):
        super().__init__()
        self.statements = tokens

    def __repr__(self):
        ret = "\n".join(repr(tok) for tok in self.statements)
        return ret


class ITE(Statement):
    """An if-then-else expression"""
    def __init__(self, tokens):
        super().__init__()
        self.node_shape = "diamond"
        self.node_classes = ["branch"]
        try:
            self.condition, self.consequence, self.alternative, *empty = tokens
            assert [] == empty
        except ValueError:
            self.condition, self.consequence, *empty = tokens
            assert [] == empty
            self.alternative = None

    def _set_id(self, id_):
        self.id = id_
        id_ += 1
        id_ = self.consequence._set_id(id_)
        if self.alternative:
            id_ = self.alternative._set_id(id_)
        return id_

    def __repr__(self):
        ret = "if ({}) {{\n{}\n}}".format(
            repr(self.condition), repr(self.consequence))
        if self.alternative:
            ret += " else {{ \n {} \n }}".format(self.alternative)
        return ret

    def simplify(self, tt, replacement_counter):
        assignments, replacements = self.condition.extract_global_read(tt, replacement_counter)
        self.condition.replace_globals(replacements)
        consequence = self.consequence.simplify(tt, replacement_counter)
        assert consequence
        if len(consequence) > 1:
            self.consequence = MultiStatement(consequence)
        else:
            self.consequence = consequence[0]
        # set label for consequence to make distinction between consequence and
        # alternative easier in graph presentation
        self.consequence.node_classes.append("true_suc")
        if isinstance(self.consequence, MultiStatement):
            if self.consequence.statements:
                self.consequence.statements[0].node_classes.append("true_suc")
        if self.alternative:
            alternative = self.alternative.simplify(tt, replacement_counter)
            assert alternative
            if len(alternative) > 1:
                self.alternative = MultiStatement(alternative)
            else:
                self.alternative = alternative[0]
            self.alternative.node_classes.append("false_suc")
            if isinstance(self.alternative, MultiStatement):
                if self.alternative.statements:
                    self.alternative.statements[0].node_classes.append("false_suc")
        assignments.append(self)
        return assignments

    def get_cfg_exit_nodes(self):
        ret = self.consequence.get_cfg_exit_nodes()
        if self.alternative:
            ret += self.alternative.get_cfg_exit_nodes()
        return ret

    def create_node_data(self, cfg_predecessors):
        ret = super().create_node_data(cfg_predecessors)
        ret += self.consequence.create_node_data([self])
        if self.alternative:
            ret += self.alternative.create_node_data([self])
        return ret

    def create_label(self):
        return self.condition.create_label()


class PrintlnStatement(Statement):
    """A println statement"""
    def __init__(self, tokens):
        super().__init__()
        self.expressions = tokens

    def __repr__(self):
        return 'println({});'.format(", ".join(repr(tok) for tok in self.expressions))

    def get_cfg_exit_nodes(self):  # pylint: disable=useless-super-delegation
        return super().get_cfg_exit_nodes()

    def create_node_data(self, cfg_predecessors):  # pylint: disable=useless-super-delegation
        return super().create_node_data(cfg_predecessors)

    def simplify(self, tt, replacement_counter):
        simplified = []
        for expression in self.expressions:
            reads, replacements = expression.extract_global_read(tt, replacement_counter)
            expression.replace_globals(replacements)
            simplified += reads
        simplified.append(self)
        return simplified


class TypeTable():
    """TypeTable is used to keep track of variables, their types and whether they are global"""
    def __init__(self):
        self._environment: typing.List[typing.Dict[str, PseuCoType]] = [{}]

    @property
    def current_env(self) -> Mapping[str, PseuCoType]:
        """The currently active environment"""
        return self._environment[-1]

    def start_fresh_environment(self):
        """Opens a new scope"""
        self._environment.append({})

    def close_fresh_environment(self) -> Mapping[str, PseuCoType]:
        """Closes the most recent scope, and returns its mapping"""
        return self._environment.pop()

    def was_declared(self, variable_name: str) -> bool:
        """Checks whether a variable was declared"""
        for env in reversed(self._environment):
            if variable_name in env:
                return True
        return False

    def is_local_variable(self, variable_name):
        """Returns true if a variable is not global"""
        if len(self._environment) == 1:
            return False
        from collections import ChainMap
        return variable_name in ChainMap(*self._environment[1:])

    def lookup_type(self, variable_name: str):
        """Checks the type of a given variable"""
        for env in reversed(self._environment):
            if variable_name in env:
                return env[variable_name]
        raise RuntimeError(("cannot use variable {}"
                            " before declaration").format(variable_name))


class VariableDeclaration(Statement):
    """A variable declaration. The initial value is optional"""
    def __init__(self, s, l, toks, *, type_table):  # pylint: disable=unused-argument
        super().__init__()
        var_type, name, *initial_value = toks
        self.var_type: PseuCoType = var_type
        self.name: str = name
        if initial_value:
            # because of the optional nature of the initial value we get a list
            # of tokens. It should contain exactly one element, the expression
            # corresponding to the initial value
            assert len(initial_value) == 1
            self.initial_value: BaseExpression = initial_value[0]
        else:
            self.initial_value: BaseExpression =\
                self.var_type.get_default_value()

        type_table.current_env[name] = var_type
        if len(self.initial_value.read_global_variables) == 1:
            self.node_classes += [
                "global_read_{}".format(self.initial_value.read_global_variables[0])
            ]

    def simplify(self, tt, replacement_counter):
        if isinstance(self.initial_value, VariableUsage):
            # do not attempt to simplfy expression when it already is a simple
            # read of a global
            replacement_counter.global_operations += 1
            return [self]
        reads, replacements = self.initial_value.extract_global_read(tt, replacement_counter)
        self.initial_value.replace_globals(replacements)
        reads.append(self)
        return reads

    def __repr__(self):
        ret = self.var_type.value + " " + self.name + " = "
        ret += repr(self.initial_value) + ";"
        return ret

    def get_cfg_exit_nodes(self):  # pylint: disable=useless-super-delegation
        return super().get_cfg_exit_nodes()

    def create_node_data(self, cfg_predecessors):  # pylint: disable=useless-super-delegation
        return super().create_node_data(cfg_predecessors)


class VariableAssignment(Statement):
    """A variable assignment"""
    def __init__(self, s, l, toks, *, type_table):
        super().__init__()
        self.var_name, self.assignment_expression, *empty = toks
        assert [] == empty
        assert type_table.was_declared(self.var_name)
        self.is_global_write = not type_table.is_local_variable(
            self.var_name)

        if self.is_global_write:
            self.node_classes.append("global_write_"+self.var_name)
        else:
            raise UnsupportedPseuCoException(
                s, l,
                "local variable {} must be constant!".format(self.var_name))

    def simplify(self, tt, replacement_counter):
        reads, replacements = self.assignment_expression.extract_global_read(tt, replacement_counter)
        self.assignment_expression.replace_globals(replacements)
        reads.append(self)
        if self.is_global_write:
            replacement_counter.global_operations += 1
        return reads

    def __repr__(self):
        return self.var_name + " = " + repr(self.assignment_expression)

    def get_cfg_exit_nodes(self):  # pylint: disable=useless-super-delegation
        return super().get_cfg_exit_nodes()

    def create_node_data(self, cfg_predecessors):  # pylint: disable=useless-super-delegation
        return super().create_node_data(cfg_predecessors)


class Initialization(ASTNode):
    """The initialization section of a TinyPseuCo program. Contains global variable declarations and
    agent definitions"""
    def __init__(self, tokens):
        super().__init__()
        self.variable_declarations = []
        self.agent_declarations = {}
        self.node_classes = []
        for token in tokens:
            if isinstance(token, VariableDeclaration):
                self.variable_declarations.append(token)
                self.node_classes.append("global_write_%s" % format(token.name))
            elif isinstance(token, SimpleAgent):
                self.agent_declarations[token.agentName] = token
            else:
                print("Unexpected element: ", token)
                print(type(token))

    def create_label(self):
        """Create a label"""
        return "\n".join((repr(declaration) for declaration in self.variable_declarations))

    def create_node_data(self) -> List[Mapping[str, Any]]:
        """Creates a serializable representation of the node to send to the JS world"""
        ret = {
            "id": 0,
            "node_label": self.create_label(),
            "node_class": " ".join(self.node_classes),
            "node_shape": "rect",
            "parents":  []
        }
        return [ret]


def createParser():  # pylint: disable=too-many-locals
    """Creates a parser for TinyPseuCo
    The parser is stateful and must be freshly created to parse a new program
    """
    tt = TypeTable()

    integer = Word(nums).setParseAction(IntegerConstant)
    identifier = ~(Literal("if") | "while" | "else"
                   | "println" | "bool" | "string" | "int") + Word(
                       alphas,
                       pp.alphanums + "_").setParseAction(lambda t: str(t[0]))
    boolean = Literal("true") | Literal("false")
    boolean.setParseAction(BooleanConstant)
    string_literal = pp.QuotedString('"').setParseAction(StringLiteral)

    constant = integer | boolean | string_literal
    operand = constant | identifier.copy(
    ).setParseAction(lambda s, l, tokens: VariableUsage(s, l, tokens, type_table=tt))

    signop = oneOf('+ -')
    multop = oneOf('* /')
    plusop = oneOf('+ -')

    arithexpression = infixNotation(operand, [
        (signop, 1, opAssoc.RIGHT, expression_list2_unary_expression),
        (multop, 2, opAssoc.LEFT, expression_list2_binary_expression),
        (plusop, 2, opAssoc.LEFT, expression_list2_binary_expression),
    ])

    compareop = oneOf('< <= >= >')
    eqop = oneOf('== !=')

    compareexpression = infixNotation(arithexpression, [
        (compareop, 2, opAssoc.LEFT, expression_list2_binary_expression),
        (eqop, 2, opAssoc.LEFT, expression_list2_binary_expression),
    ])

    booleanexpression = infixNotation(compareexpression, [
        (Literal('!'), 1, opAssoc.RIGHT, expression_list2_unary_expression),
        (Literal('||'), 2, opAssoc.LEFT, expression_list2_binary_expression),
        (Literal('&&'), 2, opAssoc.LEFT, expression_list2_binary_expression),
    ])

    expression = booleanexpression

    assignment = identifier + Suppress("=") + expression
    assignment.setParseAction(lambda s, l, t: VariableAssignment(s, l, t, type_table=tt))

    statement = pp.Forward()

    ite = (Suppress(Literal("if (")) + expression + Suppress(")") + statement +
           Optional(Suppress(Literal("else")) + statement)).setParseAction(ITE)

    printlnStatement = (
        Suppress("println(") +
        delimitedList(expression).setParseAction(PrintlnStatement) + Suppress(");"))

    typeIdentifier = Literal("int") | "bool" | "string"
    typeIdentifier.setParseAction(lambda tokens: PseuCoType(tokens[0]))

    variableDeclaration = typeIdentifier + identifier + Optional(
        Suppress("=") + expression) + Suppress(";")
    variableDeclaration.setParseAction(lambda s, l, t: VariableDeclaration(s, l, t, type_table=tt))

    statement << ((expression + Suppress(";")).setParseAction(SimpleStatement)  # pylint: disable=expression-not-assigned
                  | ite
                  | variableDeclaration
                  | pp.Suppress("{") + pp.ZeroOrMore(statement).setParseAction(
                      MultiStatement) + pp.Suppress("}")
                  | assignment + Suppress(";")
                  | printlnStatement
                  | Suppress(";").setParseAction(EmptyStatement)
                 )

    startAgent = Literal("start(") + identifier.setResultsName(
        "startedAgentNames", True) + "()" + ");"

    functionBody = ZeroOrMore(statement)
    functionBody.setParseAction(FunctionBody)

    voidFunction = (Suppress("void") + identifier + Suppress("()") + Suppress(
        "{").setParseAction(lambda t: tt.start_fresh_environment()) +
                    functionBody + Suppress("}"))
    voidFunction.setParseAction(lambda s, l, t: SimpleAgent(s, l, t, type_table=tt))

    mainAgent = (
        Suppress("mainAgent {") + ZeroOrMore(startAgent) + Suppress("}"))

    initialization = ZeroOrMore(
        (variableDeclaration
         | voidFunction))
    initialization.setParseAction(Initialization)

    program = (initialization.setResultsName("initialization") +
               mainAgent).ignore(pp.cppStyleComment())

    return program
