"""
Contains functionality to transform TinyPseuCo into z3 formulas
"""

from functools import singledispatch, update_wrapper
from itertools import chain
import typing
from typing import List, Tuple
from enum import Enum, auto, unique

import z3
from z3 import ForAll, And, Or, Implies, Exists, Not

from pseuco_parser import (BaseExpression, UnaryExpression, BinaryExpression, UnaryOperator,
                           BinaryOperator, IntegerConstant, BooleanConstant, VariableUsage,
                           Statement, EmptyStatement, SimpleStatement, MultiStatement, ITE,
                           VariableAssignment, VariableDeclaration, PrintlnStatement, PseuCoType)


@unique
class MemoryModel(Enum):
    """Enum to select which memory model to use"""
    SequentialConsistency = auto()
    WeakConsistency = auto()



def methoddispatch(func):
    """ Taken from
    https://stackoverflow.com/questions/24601722/how-can-i-use-functools-singledispatch-with-instance-methods
    Allows single dispatch for methods
    """
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):  # pylint: disable=missing-docstring
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)
    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


class Z3Compiler:

    def __init__(self, timeout=60000):
        self.context = z3.Context()
        self.local_var_table: typing.Mapping[str, z3.ArithRef] = {}
        self.global_var_table: typing.Mapping[str, "PseuCoVar"] = {}
        self.solver: z3.Solver = z3.Solver(ctx=self.context)
        self.solver.set
        self.statement_enablers: List[z3.z3.ExprRef] = []
        self.global_reads: List[Tuple["Instruction", "PseuCoVar", z3.IntNumRef]] = []
        self.global_writes: List[Tuple["Instruction", "PseuCoVar", z3.IntNumRef]] = []
        self.parent_instructions: List["Instruction"] = []
        self.current_agent_id = 0
        self.direct_predecessors = {}
        self.PseuCoVar = z3.DeclareSort('PseuCoVar', ctx=self.context)
        self.Instruction = None
        self.InstructionOrder = None
        self.Reachable = None
        self.ReadValue = None
        self.WriteValue = None

        self._instruction2statement = {}
        self._current_agent_used_id_counter = 0
        self._timeout = timeout


    def make_global_read(self, varname: str,
                         instr: "Instruction") -> z3.z3.ArithRef:
        """Takes the variable name and where it is written
        :returns: a variable reperesenting the result of the read
        """
        read_value = z3.Const(str(self.current_agent_id) + str(instr),
                              z3.IntSort(ctx=self.context))
        variable = self.global_var_table[varname]
        self.global_reads.append((instr, variable, read_value))
        return read_value

    def make_global_write(self,
                          varname: str,
                          instr: "Instruction",
                          written_value: z3.z3.ArithRef):
        """Prepares the addition of an global variable read constraint"""
        variable = self.global_var_table[varname]
        self.global_writes.append((instr, variable, written_value))

    def create_global_pseuco_variable(self, var_name):
        var = z3.Const(var_name, self.PseuCoVar)
        self.global_var_table[var_name] = var
        return var

    def create_local_pseuco_variable(self, name, var_type: PseuCoType):
        if var_type is PseuCoType.Int:
            new_var = z3.Int(str((self.current_agent_id, name)), ctx=self.context)
        elif var_type is PseuCoType.Boolean:
            new_var = z3.Bool(str((self.current_agent_id, name)), ctx=self.context)
        else:
            raise NotImplementedError("No support for %s" % str(var_type))
        assert name not in self.local_var_table, "no shadowing!"
        self.local_var_table[name] = new_var
        return new_var

    @methoddispatch
    def compile_expression(self, expression: BaseExpression) -> z3.z3.ExprRef:
        raise NotImplementedError("This does not work")

    @compile_expression.register(UnaryExpression)
    def _(self, unary_exp: UnaryExpression):
        if unary_exp.operator is UnaryOperator.Plus:
            return self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Minus:
            return -self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Not:
            return not self.compile_expression(unary_exp.operand)
        else:
            raise NotImplementedError("%s is not supported" %
                                      unary_exp.operator)

    @compile_expression.register(IntegerConstant)
    def _(self, c: IntegerConstant) -> z3.z3.IntNumRef:
        return z3.IntVal(c.value, ctx=self.context)

    @compile_expression.register(BooleanConstant)
    def _(self, c: BooleanConstant) -> z3.z3.BoolRef:
        return z3.BoolVal(c.value, ctx=self.context)

    @compile_expression.register(BinaryExpression)
    def _(self, binary_exp: BinaryExpression):
        lhs = self.compile_expression(binary_exp.lhs)
        rhs = self.compile_expression(binary_exp.rhs)
        op = binary_exp.operator
        if op is BinaryOperator.Plus:
            return lhs + rhs
        elif op is BinaryOperator.Minus:
            return lhs - rhs
        elif op is BinaryOperator.Mult:
            return lhs * rhs
        elif op is BinaryOperator.Div:
            return lhs / rhs
        elif op is BinaryOperator.And:
            return lhs and rhs
        elif op is BinaryOperator.Or:
            return lhs or rhs
        elif op is BinaryOperator.Equal:
            return lhs == rhs
        elif op is BinaryOperator.NEqual:
            return lhs != rhs
        elif op is BinaryOperator.Lt:
            return lhs < rhs
        elif op is BinaryOperator.Lte:
            return lhs <= rhs
        elif op is BinaryOperator.Gt:
            return lhs > rhs
        elif op is BinaryOperator.Gte:
            return lhs >= rhs

    @compile_expression.register(VariableUsage)
    def _(self, var: VariableUsage):
        # TODO: how to handle the global?
        # Plan: make_global_read returns the Int where the current value of the
        # read is stored
        # Before calling compile_expresion, we update the local_var_table so
        # that varname points to the freshly read value
        return self.local_var_table[var.varname]

    def _create_and_order_instruction(self,
                                      statement: Statement) -> "Instruction":
        instr = self.Instruction.__getattribute__("a%d_%d" % (self.current_agent_id,
                                                              self._current_agent_used_id_counter))
        self._current_agent_used_id_counter += 1
        self._instruction2statement[instr] = statement
        for parent in self.parent_instructions:
            self.solver.assert_and_track(self.InstructionOrder(parent, instr),
                                         "%s < %s" % (str(parent), str(instr)))
        self.parent_instructions = [instr]
        self.solver.assert_and_track(And(*self.statement_enablers, self.context) ==
                                     self.Reachable(instr),
                                     "%s <==> reachable %s" %
                                     (str(self.statement_enablers),
                                      str(instr)
                                      )
                                     )
        return instr

    def _handle_globals_in_read(self,
                                instr: "Instruction",
                                expression: BaseExpression):
        assert len(expression.read_global_variables), \
            "After simplification, expression must contain exactly one global variable"
        for varname in expression.read_global_variables:
            val = self.make_global_read(varname, instr)
            self.local_var_table[varname] = val

    @methoddispatch
    def compile_statement(self, statement: Statement):
        raise NotImplementedError("no support for %s" % str(statement))

    @compile_statement.register(EmptyStatement)
    def _(self, statement: EmptyStatement):
        return

    @compile_statement.register(SimpleStatement)
    def _(self, statement: SimpleStatement):
        assert not statement.expression.read_global_variables, \
            "simple statement must not contain globals after simplification"

    @compile_statement.register(PrintlnStatement)
    def _(self, println: PrintlnStatement):
        assert not any(expression.read_global_variables for expression in println.expressions), \
            "println statement must not contain global reads after simplification"

    @compile_statement.register(MultiStatement)
    def _(self, statement: MultiStatement):
        for statement in statement.statements:
            self.compile_statement(statement)

    @compile_statement.register(ITE)
    def _(self, ite: ITE):
        assert not ite.condition.read_global_variables, \
            "Condition must not contain global reads after simplification"
        expr = self.compile_expression(ite.condition)
        self.statement_enablers.append(expr)
        old_parents = self.parent_instructions
        consequence_parents = []
        alternative_parents = []
        self.compile_statement(ite.consequence)
        consequence_parents = self.parent_instructions
        self.parent_instructions = old_parents
        if ite.alternative:
            # flip last condition
            self.statement_enablers[-1] = Not(self.statement_enablers[-1])
            self.compile_statement(ite.alternative)
            alternative_parents = self.parent_instructions
        self.statement_enablers.pop()
        self.parent_instructions = consequence_parents + alternative_parents

    @compile_statement.register(VariableDeclaration)
    def _(self, vardec: VariableDeclaration):
        # WARNING: This only handles local variable declarations
        var = self.create_local_pseuco_variable(vardec.name, vardec.var_type)
        if (vardec.initial_value.read_global_variables):
            instr = self._create_and_order_instruction(vardec)
            self._handle_globals_in_read(instr, vardec.initial_value)
        value = self.compile_expression(vardec.initial_value)
        fact = var == value
        # self.solver.add(fact)
        self.solver.assert_and_track(fact, str(fact))

    @compile_statement.register(VariableAssignment)
    def _(self, assignment: VariableAssignment):
        assert assignment.is_global_write, "Unsupported modification of local"
        # TODO: cannot handle read and write of global in same statement
        assert not assignment.assignment_expression.read_global_variables, \
            "assignment must not contain global reads after simplification"
        written_value = self.compile_expression(
            assignment.assignment_expression)
        instr = self._create_and_order_instruction(assignment)
        self.make_global_write(assignment.var_name, instr, written_value)

    def _track_predecessors(self):
        for instruction in self._instruction2statement:
            predecessor = z3.Const("pre"+str(instruction), self.Instruction)
            pre, notpre = z3.Consts("pre notpre", self.Instruction)
            clause = Implies(
                self.Reachable(instruction),
                Or(
                    Not(Exists([notpre],
                                  self.InstructionOrder(notpre, instruction))),
                    And(
                        self.InstructionOrder(pre, instruction),
                        Exists([pre],
                               And(
                                   predecessor == pre,
                                   Not(Exists([notpre],
                                                 And(
                                                     self.Reachable(notpre),
                                                     notpre != pre,
                                                     self.InstructionOrder(notpre, instruction),
                                                     self.InstructionOrder(pre, notpre),
                                                 self.context)
                                                )
                                         ),
                               self.context)
                              ),
                    self.context),
                self.context)
            )
            self.solver.add(clause)
            self.direct_predecessors[instruction] = predecessor

    def compile(self, program):
        # 1. create the Instruction datatype/enum . It contains one constructor per instruction,
        # that is per statement which does a global read or write. Those consist of every variable
        # declaration in the initialization section, as well as writes and reads in each started
        # agent
        self.Instruction = z3.Datatype("Instruction", ctx=self.context)
        for declaration_counter in range(len(program.initialization.variable_declarations)):
            self.Instruction.declare("a0_%d" % declaration_counter)
        startedAgents = program.startedAgentNames.asList()
        for agent_counter, agent_name in enumerate(chain.from_iterable(startedAgents)):
            agent = program.initialization.agent_declarations[agent_name]
            for statement_counter in range(agent.global_operations):
                self.Instruction.declare("a%d_%d" % (agent_counter+1, statement_counter))
        self.Instruction = self.Instruction.create()

        # 2. Now that Instructions do exist, define relations on them
        self.InstructionOrder = z3.Function("InstructionOrder", self.Instruction, self.Instruction,
                                            z3.BoolSort(ctx=self.context))
        # Todo: support more than int for read and write
        self.ReadValue = z3.Function("read", self.Instruction, self.PseuCoVar,
                                     z3.IntSort(ctx=self.context), z3.BoolSort(ctx=self.context))
        self.WriteValue = z3.Function("write", self.Instruction, self.PseuCoVar,
                                      z3.IntSort(ctx=self.context), z3.BoolSort(ctx=self.context))

        # Many properties only have to hold for statements which are actually reachable
        # thus we need to define a predicate for reachability
        self.Reachable = z3.Function("reachable", self.Instruction, z3.BoolSort(ctx=self.context))

        # 3. generate constraints for global variables and their initialization
        for counter, declaration in enumerate(program.initialization.variable_declarations):
            self.create_global_pseuco_variable(declaration.name)
            instr = self.Instruction.__getattribute__("a0_%d" % counter)
            self._instruction2statement[instr] = declaration
            self.solver.assert_and_track(self.Reachable(instr), "reach %s" % str(instr))
            self.parent_instructions.append(instr)
            expr = self.compile_expression(declaration.initial_value)
            self.make_global_write(declaration.name, instr, expr)
        initial_parents = self.parent_instructions

        # 4. Iterate over all started agents, and generate code for them
        for agent_name in chain.from_iterable(startedAgents):
            self.parent_instructions = initial_parents
            self.current_agent_id += 1
            self._current_agent_used_id_counter = 0
            self.local_var_table = {}  # reset local variables
            agent = program.initialization.agent_declarations[agent_name]
            body = agent.functionBody
            for statement in body.statements:
                self.compile_statement(statement)
        a = z3.Const('a', self.Instruction)
        x = z3.Const('x', self.PseuCoVar)
        p = z3.Const('p', z3.IntSort(ctx=self.context))
        # specify read and write function
        reads = ForAll([a, x, p],
                       self.ReadValue(a, x, p) == (
                           Or(*(And(a == b, x == y, p == q, self.context)
                                for (b, y, q)
                                in self.global_reads), self.context))
                       )
        writes = ForAll([a, x, p],
                        self.WriteValue(a, x, p) == (
                            Or(*(And(a == b, x == y, p == q, self.context)
                                 for (b, y, q)
                                 in self.global_writes), self.context))
                        )
        self.solver.assert_and_track(reads, "reads")
        self.solver.assert_and_track(writes, "writes")
        self._track_predecessors()
        sat_stat = self.solver.check()
        assert sat_stat == z3.sat, sat_stat

    def check_sat(self, memory_model: MemoryModel):
        self.solver.push()
        if memory_model is MemoryModel.SequentialConsistency:
            self.add_seqconsist_axioms()
        elif memory_model is MemoryModel.WeakConsistency:
            self.add_weak_axioms()
        is_sat = self.solver.check()
        if is_sat == z3.unknown:
            if self.solver.reason_unknown() == 'timeout':
                raise ResourceWarning("Timeout occured")
            self.solver.pop()
            return "Cannot decide given program. WTH did you do?"
        if is_sat == z3.unsat:
            reply = []
            reply.append(100*"_")
            reply.append("UNSAT")
            reply.append(100*"^")
            reply.append(self.solver.unsat_core())
            self.solver.pop()
            return "\n".join(reply)
        counter = 0
        reply = []
        while is_sat == z3.sat and counter < 50:
            counter += 1
            model = self.solver.model()
            reply.append("solution number %d" % counter)
            for (instr, _, value) in self.global_reads:
                text = repr(self._instruction2statement[instr])
                reply.append(text + ": " + str(model[value].as_long()))
            reply.append(100*"-")
            current_model = []
            for (instr, variable, value) in self.global_reads:
                current_model.append(value == model[value].as_long())
            self.solver.assert_and_track(Not(And(*current_model, self.context)),
                                         "counter %s" % counter,)
            is_sat = self.solver.check()
        self.solver.pop()
        if counter == 50:
            reply = ["Too many possible reads. Potential thin air read?"]
        return "\n".join(reply)

    def add_seqconsist_axioms(self):
        """Creates the axioms of sequential consistency"""
        a, b, c = z3.Consts('a b c', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self.context))

        seqconsist_axioms = {
            "total": ForAll([a, b], Implies(
                And(self.Reachable(a), self.Reachable(b), self.context),
                Or(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self.context)
            )),  # order is total on reachable statements
            "transitive": ForAll([a, b, c], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, c), self.context),
                self.InstructionOrder(a, c)
            )),  # transitive
            "antisymmetric": ForAll([a, b], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self.context),
                a == b
            )),
            "write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self.context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                self.InstructionOrder(b, a),
                                Not(
                                    Exists([c, q], And(
                                        self.Reachable(c),
                                        c != b,
                                        self.InstructionOrder(b, c),
                                        self.InstructionOrder(c, a),
                                        self.WriteValue(c, x, q),
                                        self.context
                                    ))
                                ),
                                self.context
                                ))
            )),
        }
        for name, fact in seqconsist_axioms.items():
            self.solver.assert_and_track(fact, name)

    def add_weak_axioms(self):
        """Creates the axioms of sequential consistency"""
        a, b, c = z3.Consts('a b c', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self.context))


        weak_order_axioms = {
            "reflexive": ForAll([a] , Implies(self.Reachable(a),
                                self.InstructionOrder(a, a))),  # order is reflexive
            "transitive": ForAll([a, b, c], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, c), self.context),
                self.InstructionOrder(a, c)
            )),  # transitive
            "antisymmetric": ForAll([a, b], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self.context),
                a == b
            )),
            "write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self.context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                Not(self.InstructionOrder(a, b)),
                                Not(
                                    Exists([c, q], And(
                                        self.Reachable(c),
                                        c != b,
                                        self.InstructionOrder(b, c),
                                        self.InstructionOrder(c, a),
                                        self.WriteValue(c, x, q),
                                        self.context
                                    ))
                                ),
                                self.context
                                ))
            )),
        }
        for name, fact in weak_order_axioms.items():
            self.solver.assert_and_track(fact, name)
