"""The m3 server, which handles all requests over WebSockets and delegates the actual work to the
pseucparser and to the Z3Compiler"""

import itertools
import time
import json
import asyncio
import multiprocessing
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor

from autobahn.asyncio.websocket import WebSocketServerProtocol, \
    WebSocketServerFactory

from pseuco_parser import createParser, UnsupportedPseuCoException
from semantic import Z3Compiler, MemoryModel


def get_cfg_for_program(program):
    """Creates a cfg representation of the given program"""
    ret = {}
    counter = 0
    for agent_name in itertools.chain.from_iterable(
            program.startedAgentNames.asList()):
        agent = program.initialization.agent_declarations[agent_name]
        body = agent.functionBody
        cfg_predecessors = []
        cfc_graph_representation = []
        ret["init0"] = program.initialization.create_node_data()
        for statement in body.statements:
            cfc_graph_representation += statement.create_node_data(
                cfg_predecessors)
            cfg_predecessors = statement.get_cfg_exit_nodes()
        ret[agent_name + str(counter+1)] = cfc_graph_representation
        counter += 1
    return ret


class BookKeeper:
    """Keeps track of how many ressources a client is using,
       Currently broken, and thus not used
    """
    def __init__(self, score_limit=1000):
        """Sets up the bookkeeping"""
        self.ip2score = defaultdict(int)
        self.ip2last_access = {}
        self.time_of_last_reset = time.monotonic()
        self.limit = score_limit

    def _improve_score(self):
        time_since_reset = time.monotonic() - self.time_of_last_reset
        score_reduction = int(time_since_reset) / (60000 * 30)  # 15 minutes
        if time_since_reset <= 0:
            return
        self.time_of_last_reset = time.monotonic()
        improvement = int(time_since_reset)
        removable_users = []
        for ip, score in self.ip2score.items():
            new_score = max(0, score - 100*score_reduction)
            if new_score == 0:
                removable_users.append(ip)
            else:
                self.ip2score[ip] = new_score
        for user in removable_users:
            del self.ip2last_access[ip]
            del self.ip2score[ip]

    def caused_timeout(self, ip):
        """Called when a user caused a timeout of the z3 solver"""

    def track_and_check_visit(self, ip):
        """Call when user visits site"""
        self._improve_score()
        self.ip2last_access[ip] = time.monotonic()
        return self.ip2score[ip] > self.limit

    def update_score(self, ip):
        last_access = time.monotonic()
        time_spent = int(self.ip2last_access[ip] - last_access)
        self.ip2score[ip] += (time_spent+1)*50

book_keeper = BookKeeper()


class M3ServerProtocoll(WebSocketServerProtocol):

    executor = ThreadPoolExecutor(max_workers=multiprocessing.cpu_count(),
                                  thread_name_prefix="z3_task")

    def __init__(self):
        super().__init__()
        self.compiler = Z3Compiler()

    def onConnect(self, request):
        print("Client connecting: {0}".format(request.peer))

    def onOpen(self):
        print("Websocket connection is open.")

    async def onMessage(self, payload, isBinary):
        assert not isBinary, "Where did the binary message come from" #  turn off in production
        decoded_message = json.loads(payload.decode('utf8'))
        if decoded_message["type"] == "code":
            try:
                tinyPseucoParser = createParser()
                program = tinyPseucoParser.parseString(decoded_message["code"])
                reply = {
                    "cfg": get_cfg_for_program(program),
                    "type": "parseResult"
                }
                loop = asyncio.get_event_loop()
                await loop.run_in_executor(
                    M3ServerProtocoll.executor,
                    self.compiler.compile,
                    program,
                )
            except UnsupportedPseuCoException as e:
                reply = {
                    "type": "error",
                    "errorMessage": str(e),
                }
        elif decoded_message["type"] == "listAllReads":
            mem_model = {
                "seqConsist": MemoryModel.SequentialConsistency,
                "weak": MemoryModel.WeakConsistency
            }
            selected_model = decoded_message["memoryModel"]
            loop = asyncio.get_event_loop()
            allReads = await loop.run_in_executor(
                M3ServerProtocoll.executor,
                self.compiler.check_sat,
                mem_model[selected_model],
            )
            allReads =  self.compiler.check_sat(mem_model[selected_model])
            reply = {
                "type": "allReads",
                "allReads": allReads
            }
        elif decoded_message["type"] == "WriteSeenAdded":
            print(decoded_message)
            reply = {
                "type": "NotImplemented"
            }

        self.sendMessage(json.dumps(reply).encode('utf8'))

    def onClose(self, wasClean, code, reason):
        print("WebSocket connection closed: {0}".format(reason))
        print("wasClean: {} code:{}".format(wasClean, code))


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="The m3 server")
    parser.add_argument("--port", "-p", type=int, help="The port, defaults to 5678", default=5678)
    parser.add_argument("--hostname", "-n", type=str, help="The hostname, defaults to 127.0.0.1",
                        default="127.0.0.1")
    args = parser.parse_args()

    host = args.hostname
    port = args.port

    factory = WebSocketServerFactory("ws://%s:%d" % (host, port))
    factory.protocol = M3ServerProtocoll

    loop = asyncio.get_event_loop()
    coro = loop.create_server(factory, host, port)

    server = loop.run_until_complete(coro)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.close()
        loop.close()
